from decimal import Decimal
from typing import Protocol

from .building import Building
from .tenant import Tenant
from .time_period import TimePeriod


class AllocationStrategy(Protocol):
    """
    A class which calculates how costs are allocated among tenants.
    """

    @property
    def name(self) -> str: ...

    def total_shares(
        self, accounting_period: TimePeriod, building: Building
    ) -> Decimal: ...

    def tenant_shares(
        self, accounting_period: TimePeriod, building: Building, tenant: Tenant
    ) -> Decimal: ...



from dataclasses import dataclass
from decimal import Decimal
from typing import Collection, Final, Iterator

from .invoice import Invoice


@dataclass(frozen=True, slots=True)
class InvoiceCollection:
    name: str
    invoices: Collection[Invoice]

    @property
    def gross_total(self) -> Decimal:
        """The sum of the gross amounts of all included invoices"""
        return Decimal(sum(i.gross_amount for i in self.invoices))

    @property
    def privileged_invoices(self) -> Iterator[Invoice]:
        return filter(lambda i: i.privileged_amount > 0, self.invoices)

    @property
    def privileged_invoices_by_issuer(self) -> dict[str, list[Invoice]]:
        issuer_invoices: Final[dict[str, list[Invoice]]] = dict()
        for invoice in self.privileged_invoices:
            if invoice.issuer_name in issuer_invoices:
                issuer_invoices[invoice.issuer_name].append(invoice)
            else:
                issuer_invoices[invoice.issuer_name] = [invoice]
        return issuer_invoices



from dataclasses import dataclass
from decimal import Decimal


@dataclass(frozen=True, slots=True)
class LaborCostItem:
    collection_name: str
    issuer_name: str
    gross_amount: Decimal
    privileged_amount: Decimal
    factor: Decimal
    share_amount: Decimal



import datetime as dt
from dataclasses import dataclass, field
from decimal import Decimal


@dataclass(frozen=True, slots=True)
class Invoice:
    issuer_name: str
    date: dt.date
    gross_amount: Decimal
    privileged_amount: Decimal = field(default_factory=lambda: Decimal("0.00"))
    net_amount: Decimal | None = None
    name: str | None = None

    def __post_init__(self) -> None:
        self._validate_gross_not_less_net()

    def _validate_gross_not_less_net(self) -> None:
        if not self.net_amount:
            return
        if (self.gross_amount < 0) != (self.net_amount < 0):
            raise ValueError("Net amount and gross amount must have the same sign.")
        if (abs(self.gross_amount) - abs(self.net_amount)) < 0:
            raise ValueError("Net amount cannot be greater than gross amount.")



from dataclasses import dataclass
from decimal import Decimal
from typing import Final

from nekore.models import Building, Tenant, TimePeriod


@dataclass(frozen=True, slots=True)
class ResidentNumberStrategy:
    name: str = "PNZ"

    def total_shares(
        self, accounting_period: TimePeriod, building: Building
    ) -> Decimal:
        tenants: Final = building.tenants
        total_person_days = sum(person_days(t, accounting_period) for t in tenants)
        return Decimal(total_person_days) / accounting_period.duration.days

    def tenant_shares(
        self, accounting_period: TimePeriod, _: Building, tenant: Tenant
    ) -> Decimal:
        _person_days: Final = Decimal(person_days(tenant, accounting_period))
        return _person_days / accounting_period.duration.days


def person_days(tenant: Tenant, accounting_period: TimePeriod) -> int:
    days_of_use: Final = tenant.time_of_use_in(accounting_period).days
    return tenant.number_of_people * days_of_use



from fpdf import FPDF

from .assets import top_ornament


class PDF(FPDF):
    def __init__(self) -> None:
        super().__init__()
        self.set_font("helvetica", size=11)

    def header(self) -> None:
        self.set_margin(0)
        self.image(top_ornament, w=self.epw)

    def footer(self) -> None:
        # Position cursor at 1.5 cm from bottom:
        self.set_y(-15)
        # Setting font: helvetica italic 8
        self.set_text_color(0, 0, 0)
        self.set_font("helvetica", "I", 8)
        # Printing page number:
        self.cell(0, 10, f"Seite {self.page_no()}/{{nb}}", align="C")



import datetime as dt
from dataclasses import dataclass
from typing import Final

from fpdf import FPDF

from nekore.models import Accounting, Contact

from .document_template import PDF


@dataclass(frozen=True, slots=True)
class FontSize:
    TEXT = 11
    TABLE = 10
    H1 = 16
    H2 = 14


@dataclass(frozen=True, slots=True)
class Color:
    DARKGREEN = {"r": 70, "g": 138, "b": 26}
    BLACK = {"r": 0, "g": 0, "b": 0}
    DARKRED = {"r": 200, "g": 0, "b": 0}
    GRAY = {"r": 220, "g": 220, "b": 220}


class DocumentGenerator:
    def create_documents(self, directory: str, accountings: list[Accounting]) -> None:
        for n, accounting in enumerate(accountings):
            doc = self.create_document(accounting)
            datestr = accounting.date.strftime("%Y-%m-%d")
            file = f"{datestr}_Abrechnung_{n + 1}_{accounting.recipient.name}.pdf"
            doc.output(f"{directory}/{file}")

    def format_address(self, issuer: Contact, separator: str) -> str:
        sections: Final = (
            issuer.name,
            f"{issuer.street} {issuer.house_number}",
            f"{issuer.zip_code} {issuer.city}",
        )
        return separator.join(sections)

    def create_document(self, accounting: Accounting) -> PDF:
        pdf = PDF()
        pdf.add_page()
        pdf.set_margin(10)
        # Accounting issuer
        pdf.cell(w=None, h=20, text=self.format_address(accounting.issuer, " | "))
        # Accounting date
        pdf.cell(0, 20, dt.date.today().strftime("%d.%m.%Y"), align="R")
        pdf.ln()

        pdf.multi_cell(0, 6, self.format_address(accounting.recipient, "\n"))
        pdf.ln()

        self.create_h1(
            pdf,
            "Betriebskostenabrechnung "
            + f"{accounting.accounting_period.start.strftime('%d.%m.%Y')} - "
            + accounting.accounting_period.end.strftime("%d.%m.%Y"),
        )

        self.basic_data_table(pdf, accounting)

        self.create_h2(pdf, "Umlageschlüssel")

        self.name_value_table(
            pdf,
            STK="Nach Anzahl der Wohneinheiten",
            PNZ="Nach Personenanzahl und Nutzungszeitraum",
            NHN="Nach Nutzfläche der Wohneinheiten",
        )

        self.create_h2(pdf, "Zusammenfassung")

        pdf.set_font("helvetica", "", FontSize.TEXT)
        pdf.set_text_color(**Color.BLACK)

        third_width = pdf.epw / 3
        pdf.cell(third_width, 8, "Ihre Betriebskosten", align="L")
        pdf.cell(third_width, 8, "(inkl. ges. MwSt.)", align="C")
        pdf.cell(
            third_width,
            8,
            f"{accounting.gross_total:.2f} EUR".replace(".", ","),
            align="R",
        )
        pdf.ln()

        pdf.cell(third_width, 8, "Ihre Vorauszahlung", align="L")
        pdf.cell(third_width, 8, "(inkl. ges. MwSt.)", align="C")
        pdf.cell(
            third_width, 8, f"{accounting.prepaid:.2f} EUR".replace(".", ","), align="R"
        )
        pdf.ln()

        pdf.set_fill_color(**Color.GRAY)
        pdf.set_text_color(
            **Color.DARKGREEN if accounting.has_refund else Color.DARKRED
        )
        pdf.set_font("helvetica", "B", FontSize.TEXT)
        pdf.cell(
            third_width,
            8,
            "Ihre Rückzahlung" if accounting.has_refund else "Ihre Nachzahlung",
            fill=True,
        )
        pdf.cell(
            third_width,
            8,
            "(inkl. ges. MwSt.)",
            fill=True,
            align="C",
        )
        pdf.cell(
            third_width,
            8,
            f"{abs(accounting.refund):.2f} EUR".replace(".", ","),
            fill=True,
            align="R",
        )
        pdf.ln()

        pdf.set_font("helvetica", "", FontSize.TEXT)
        pdf.set_text_color(**Color.BLACK)
        if accounting.has_refund:
            pdf.cell(
                0,
                15,
                "Sie erhalten Ihre Rückzahlung innerhalb der nächsten 28 Tage als "
                "Überweisung auf Ihr Konto.",
            )
        else:
            pdf.cell(
                0,
                15,
                "Bitte überweisen Sie den Nachzahlungsbetrag innerhalb der "
                "nächsten 28 Tage.",
            )
            pdf.ln()
            if accounting.issuer.bank_account is not None:
                pdf.cell(40, 8, f"{accounting.issuer.name}")
                pdf.cell(40, 8, f"{accounting.issuer.bank_account.bank}")
                pdf.cell(60, 8, f"{accounting.issuer.bank_account.iban}")
                pdf.cell(0, 8, f"{accounting.issuer.bank_account.bic}")

        pdf.add_page()
        pdf.set_margin(10)
        self.create_h2(pdf, "Umlage Nebenkosten")
        self.create_allocation_table(pdf, accounting)

        self.create_h2(pdf, "Ausweis von Arbeitskosten nach § 35a EStG")
        self.create_labor_cost_table(pdf, accounting)
        return pdf

    def create_h1(self, pdf: FPDF, text: str) -> None:
        pdf.set_fill_color(**Color.DARKGREEN)
        pdf.cell(pdf.epw, 0.7, "", new_y="NEXT", new_x="LMARGIN", fill=True)
        pdf.ln()
        pdf.set_font("helvetica", "B", FontSize.H1)
        pdf.set_text_color(**Color.DARKGREEN)
        pdf.set_draw_color(**Color.DARKGREEN)
        pdf.cell(0, 15, text, border="T")
        pdf.ln()

    def create_h2(self, pdf: FPDF, text: str) -> None:
        pdf.set_font("helvetica", "B", FontSize.H2)
        pdf.set_text_color(**Color.DARKGREEN)
        pdf.cell(0, 12, text, new_y="NEXT", new_x="LMARGIN")

    def basic_data_table(self, pdf: FPDF, accounting: Accounting) -> None:
        basic_data: Final = {
            "Lage": accounting.apartment_name,
            "Nutzfläche": f"{accounting.floor_space:.2f} m²".replace(".", ","),
            "Personen": str(accounting.number_of_people),
            "Abrechnungszeitraum": (
                f"{accounting.accounting_period.start.strftime('%d.%m.%Y')} - "
                + accounting.accounting_period.end.strftime("%d.%m.%Y")
                + f", {accounting.accounting_period.duration.days} Tage"
            ),
            "Nutzungszeitraum": (
                f"{accounting.usage_period.start.strftime('%d.%m.%Y')} - "
                + accounting.usage_period.end.strftime("%d.%m.%Y")
                + f", {accounting.usage_period.duration.days} Tage"
            ),
        }
        self.name_value_table(pdf, **basic_data)

    def name_value_table(self, pdf: FPDF, **kwargs: str) -> None:
        pdf.set_font("helvetica", "", FontSize.TEXT)
        pdf.set_text_color(**Color.BLACK)
        for key, value in kwargs.items():
            pdf.cell(70, 6, key, align="L")
            pdf.cell(
                30,
                6,
                value,
                align="L",
                new_y="NEXT",
                new_x="LMARGIN",
            )
        pdf.ln()

    def create_allocation_table(self, pdf: FPDF, accounting: Accounting) -> None:
        # Table header
        pdf.set_text_color(**Color.BLACK)
        pdf.set_draw_color(**Color.BLACK)
        pdf.set_font("helvetica", "B", FontSize.TABLE)
        header_align = "L"
        pdf.epw
        col_w: list[int] = [55, 20, 30, 24, 30, 0]
        pdf.cell(col_w[0], 6, "Kategorie", border="B", align=header_align)
        pdf.cell(col_w[1], 6, "Schlüssel", border="B", align=header_align)
        pdf.cell(col_w[2], 6, "Anteile gesamt", border="B", align=header_align)
        pdf.cell(col_w[3], 6, "Ihre Anteile", border="B", align=header_align)
        pdf.cell(col_w[4], 6, "Gesamtkosten", border="B", align=header_align)
        pdf.cell(col_w[5], 6, "Ihr Kostenanteil", border="B", align=header_align)
        pdf.ln()
        # Table content
        pdf.set_font("helvetica", "", FontSize.TABLE)
        for n, ai in enumerate(accounting.allocation_items):
            last_row = n == len(accounting.allocation_items) - 1
            border = "B" if last_row else ""

            pdf.cell(col_w[0], 6, ai.name, border=border)
            pdf.cell(col_w[1], 6, ai.allocation_name, border=border)
            pdf.cell(
                col_w[2],
                6,
                f"{ai.shares_total:.2f}".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.cell(
                col_w[3],
                6,
                f"{ai.shares_allocated:.2f}".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.cell(
                col_w[4],
                6,
                f"{ai.gross_total:.2f} EUR".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.cell(
                col_w[5],
                6,
                f"{ai.gross_share:.2f} EUR".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.ln()
        pdf.set_font("helvetica", "", FontSize.TEXT)
        pdf.cell(
            0,
            8,
            f"Summe: {accounting.gross_total:.2f} EUR".replace(".", ","),
            align="R",
        )
        pdf.ln()
        pdf.cell(
            0,
            8,
            f"Vorauszahlung: {accounting.prepaid:.2f} EUR".replace(".", ","),
            align="R",
        )
        pdf.ln()
        pdf.set_text_color(
            **Color.DARKGREEN if accounting.has_refund else Color.DARKRED
        )
        pdf.cell(
            0,
            8,
            f"{'Ihre Rückzahlung' if accounting.has_refund else 'Ihre Nachzahlung'}: "
            f"{abs(accounting.refund):.2f} EUR".replace(".", ","),
            align="R",
        )
        pdf.ln()

    def create_labor_cost_table(self, pdf: FPDF, accounting: Accounting) -> None:
        pdf.set_text_color(**Color.BLACK)
        pdf.set_draw_color(**Color.BLACK)
        pdf.set_font("helvetica", "B", FontSize.TABLE)
        col_w: list[int] = [55, 40, 27, 27, 15, 0]
        # Table header
        header_align = "L"
        pdf.cell(col_w[0], 6, "Kategorie", border="B", align=header_align)
        pdf.cell(col_w[1], 6, "Dienstleister", border="B", align=header_align)
        pdf.cell(col_w[2], 6, "Kosten", border="B", align=header_align)
        pdf.cell(col_w[3], 6, "begünstigt", border="B", align=header_align)
        pdf.cell(col_w[4], 6, "Faktor", border="B", align=header_align)
        pdf.cell(col_w[5], 6, "Anteil", border="B", align=header_align)
        pdf.ln()
        # Table content
        pdf.set_font("helvetica", "", FontSize.TABLE)
        for n, lci in enumerate(accounting.labor_cost_items):
            last_row = n == len(accounting.labor_cost_items) - 1
            border = "B" if last_row else ""

            pdf.cell(col_w[0], 6, lci.collection_name, border=border)
            pdf.cell(col_w[1], 6, lci.issuer_name, border=border, align="R")
            pdf.cell(
                col_w[2],
                6,
                f"{lci.gross_amount:.2f} EUR".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.cell(
                col_w[3],
                6,
                f"{lci.privileged_amount:.2f} EUR".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.cell(
                col_w[4],
                6,
                f"{lci.factor:.2f}".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.cell(
                col_w[5],
                6,
                f"{lci.share_amount:.2f} EUR".replace(".", ","),
                border=border,
                align="R",
            )
            pdf.ln()
        pdf.cell(
            0,
            8,
            f"Summe: {accounting.labor_costs_total:.2f} EUR".replace(".", ","),
            align="R",
        )



from dataclasses import dataclass
from datetime import timedelta
from decimal import Decimal
from itertools import combinations
from typing import Final, Iterable, Iterator

from .contact import Contact
from .time_period import TimePeriod


@dataclass(frozen=True, slots=True)
class Tenant:
    contact: Contact
    number_of_people: int
    period: TimePeriod
    prepaid: Decimal = Decimal(0)

    def time_of_use_in(self, period: TimePeriod) -> timedelta:
        use_time: Final = self.period.intersection(period)
        if use_time is None:
            return timedelta(0)
        return use_time.duration


def overlapping_in_time(tenants: Iterable[Tenant]) -> Iterator[tuple[Tenant, Tenant]]:
    tenant_pairs: Final = combinations(tenants, 2)
    return filter(lambda t: t[0].period.overlaps(t[1].period), tenant_pairs)



from dataclasses import dataclass
from decimal import Decimal
from typing import Any

from nekore.models import Tenant


@dataclass(frozen=True, slots=True)
class DirectInvoiceStrategy:
    tenant: Tenant
    name: str = "Direkt"

    def total_shares(self, *_: Any) -> Decimal:
        return Decimal("1")

    def tenant_shares(self, _: Any, __: Any, tenant: Tenant) -> Decimal:
        return Decimal("1") if tenant == self.tenant else Decimal(0)



from dataclasses import dataclass
from itertools import chain
from typing import Collection, Final

from nekore.models import (
    Accounting,
    AllocationItem,
    Building,
    Contact,
    LaborCostItem,
    Tenant,
    TimePeriod,
)

from .invoice_collection_processor import InvoiceCollectionProcessor


@dataclass(frozen=True, slots=True)
class AccountingProcessor:
    issuer: Contact
    period: TimePeriod
    building: Building
    invoice_collection_processors: Collection[InvoiceCollectionProcessor]

    def create_accountings(self) -> list[Accounting]:
        tenants_in_period: Final = self.building.tenants_in(self.period)
        return [self.create_accounting(tenant) for tenant in tenants_in_period]

    def create_accounting(self, tenant: Tenant) -> Accounting:
        apartment: Final = self.building.apartment_of(tenant)
        return Accounting(
            issuer=self.issuer,
            recipient=tenant.contact,
            number_of_people=tenant.number_of_people,
            floor_space=apartment.floor_space,
            apartment_name=apartment.name or "",
            accounting_period=self.period,
            usage_period=tenant.period,
            prepaid=tenant.prepaid,
            allocation_items=self.create_allocation_items(tenant),
            labor_cost_items=self.create_labor_cost_items(tenant),
        )

    def create_allocation_items(self, tenant: Tenant) -> list[AllocationItem]:
        """
        Create allocation items for a tenant's accounting.
        Allocation items with a gross of zero are omitted.
        """
        allocation_items: Final = (
            processor.create_allocation_item(self.building, tenant, self.period)
            for processor in self.invoice_collection_processors
        )
        return list(filter(lambda i: i.gross_share > 0, allocation_items))

    def create_labor_cost_items(self, tenant: Tenant) -> list[LaborCostItem]:
        invoice_processors: Final = (
            p.create_labor_cost_items for p in self.invoice_collection_processors
        )
        item_groups: Final = (
            p(self.building, tenant, self.period) for p in invoice_processors
        )
        items: Final = chain(*item_groups)
        return list(filter(lambda i: i.share_amount > 0, items))



from typing import Final

__all__: Final = [
    "Accounting",
    "AllocationItem",
    "AllocationStrategy",
    "Apartment",
    "BankAccount",
    "Building",
    "Contact",
    "Invoice",
    "InvoiceCollection",
    "LaborCostItem",
    "Tenant",
    "TimePeriod",
]


from .accounting import Accounting
from .allocation_item import AllocationItem
from .allocation_strategy import AllocationStrategy
from .apartment import Apartment
from .bank_account import BankAccount
from .building import Building
from .contact import Contact
from .invoice import Invoice
from .invoice_collection import InvoiceCollection
from .labor_cost_item import LaborCostItem
from .tenant import Tenant
from .time_period import TimePeriod



import datetime as dt
from dataclasses import dataclass
from typing import Final, Self


@dataclass(frozen=True, eq=True, slots=True)
class TimePeriod:
    """A period between two dates."""

    start: dt.date
    end: dt.date

    def __post_init__(self) -> None:
        if self.start > self.end:
            raise ValueError("Start cannot be greater than end.")

    @property
    def duration(self) -> dt.timedelta:
        return self.end - self.start + dt.timedelta(days=1)

    def __contains__(self, other: dt.date | Self) -> bool:
        if isinstance(other, dt.date):
            return self.start <= other <= self.end
        return self.start <= other.start <= other.end <= self.end

    def union(self, *others: Self) -> Self:
        """Create the smallest time period that includes all given time periods."""
        all_periods: Final = others + (self,)
        total_start: Final = min(t.start for t in all_periods)
        total_end: Final = max(t.end for t in all_periods)
        return self.__class__(start=total_start, end=total_end)

    def intersection(self, *others: Self) -> Self | None:
        """
        Calculate the time period included in all periods.
        The duration includes start and end dates.
        """
        all_periods: Final = others + (self,)
        max_start: Final = max(t.start for t in all_periods)
        min_end: Final = min(t.end for t in all_periods)
        if max_start > min_end:
            return None
        return self.__class__(start=max_start, end=min_end)

    def overlaps(self, other: Self) -> bool:
        return bool(self.intersection(other))



from dataclasses import dataclass
from decimal import Decimal
from typing import Final

from nekore.models import Building, Tenant, TimePeriod


@dataclass(frozen=True, slots=True)
class FloorAreaStrategy:
    name: str = "NHN"

    def total_shares(self, _: TimePeriod, building: Building) -> Decimal:
        return Decimal(sum(a.floor_space for a in building.apartments))

    def tenant_shares(
        self, accounting_period: TimePeriod, building: Building, tenant: Tenant
    ) -> Decimal:
        apartment = building.apartment_of(tenant)
        days_of_use: Final = Decimal(tenant.time_of_use_in(accounting_period).days)
        time_share = days_of_use / accounting_period.duration.days
        return apartment.floor_space * time_share



from dataclasses import dataclass
from decimal import Decimal
from typing import Final

from nekore.models import (
    AllocationItem,
    Building,
    InvoiceCollection,
    LaborCostItem,
    Tenant,
    TimePeriod,
)
from nekore.models.allocation_strategy import AllocationStrategy


@dataclass(slots=True, frozen=True)
class InvoiceCollectionProcessor:
    invoice_collection: InvoiceCollection
    allocation_strategy: AllocationStrategy

    def total_shares(
        self, accounting_period: TimePeriod, building: Building
    ) -> Decimal:
        return self.allocation_strategy.total_shares(accounting_period, building)

    def tenant_shares(
        self, accounting_period: TimePeriod, building: Building, tenant: Tenant
    ) -> Decimal:
        return self.allocation_strategy.tenant_shares(
            accounting_period, building, tenant
        )

    def create_allocation_item(
        self, building: Building, tenant: Tenant, period: TimePeriod
    ) -> AllocationItem:
        total_shares: Final = self.total_shares(period, building)
        tenant_shares: Final = self.tenant_shares(period, building, tenant)
        ratio: Final = tenant_shares / total_shares
        return AllocationItem(
            gross_total=self.invoice_collection.gross_total,
            gross_share=ratio * self.invoice_collection.gross_total,
            shares_total=total_shares,
            shares_allocated=tenant_shares,
            allocation_name=self.allocation_strategy.name,
            name=self.invoice_collection.name,
        )

    def create_labor_cost_items(
        self, building: Building, tenant: Tenant, period: TimePeriod
    ) -> list[LaborCostItem]:
        total_shares: Final = self.total_shares(period, building)
        tenant_shares: Final = self.tenant_shares(period, building, tenant)
        ratio: Final = tenant_shares / total_shares

        issuer_invoices: Final = self.invoice_collection.privileged_invoices_by_issuer
        return [
            LaborCostItem(
                collection_name=self.invoice_collection.name,
                issuer_name=issuer,
                gross_amount=Decimal(sum(i.gross_amount for i in invoices)),
                privileged_amount=Decimal(sum(i.privileged_amount for i in invoices)),
                factor=ratio,
                share_amount=ratio * sum(i.privileged_amount for i in invoices),
            )
            for issuer, invoices in issuer_invoices.items()
        ]



import datetime as dt
from dataclasses import dataclass, field
from decimal import Decimal
from typing import Collection

from .allocation_item import AllocationItem
from .contact import Contact
from .labor_cost_item import LaborCostItem
from .time_period import TimePeriod


@dataclass(frozen=True, slots=True)
class Accounting:
    issuer: Contact
    recipient: Contact
    number_of_people: int
    floor_space: Decimal
    apartment_name: str
    accounting_period: TimePeriod
    usage_period: TimePeriod
    prepaid: Decimal
    allocation_items: Collection[AllocationItem]
    labor_cost_items: Collection[LaborCostItem]
    date: dt.date = field(default_factory=dt.date.today)

    @property
    def refund(self) -> Decimal:
        return self.prepaid - self.gross_total

    @property
    def gross_total(self) -> Decimal:
        return Decimal(sum(i.gross_share for i in self.allocation_items))

    @property
    def labor_costs_total(self) -> Decimal:
        return Decimal(sum(lc.share_amount for lc in self.labor_cost_items))

    @property
    def has_refund(self) -> bool:
        return self.refund >= 0



from dataclasses import dataclass
from decimal import Decimal


@dataclass(frozen=True, slots=True)
class AllocationItem:
    name: str
    gross_total: Decimal
    gross_share: Decimal
    shares_total: Decimal
    shares_allocated: Decimal
    allocation_name: str



from dataclasses import dataclass, field
from typing import Collection, Final, Iterator

from .apartment import Apartment
from .tenant import Tenant
from .time_period import TimePeriod


@dataclass(frozen=True, slots=True)
class Building:
    name: str | None = None
    apartments: Collection[Apartment] = field(default_factory=tuple)

    @property
    def tenants(self) -> Iterator[Tenant]:
        """All tenants from all apartments."""
        return (t for a in self.apartments for t in a.tenants)

    def tenants_in(self, period: TimePeriod) -> Iterator[Tenant]:
        """
        Get all tenants of all apartments who occupied
        an apartment during that given period.
        """
        return (t for t in self.tenants if period.overlaps(t.period))

    def apartment_of(self, tenant: Tenant) -> Apartment:
        apartments: Final = (a for a in self.apartments if tenant in a.tenants)
        try:
            return next(apartments)
        except StopIteration:
            raise ValueError(f"{tenant} must be resident in an apartment of {self}.")



from typing import Final

__all__: Final = [
    "AccountingProcessor",
    "ApartmentStrategy",
    "DirectInvoiceStrategy",
    "FloorAreaStrategy",
    "InvoiceCollectionProcessor",
    "ResidentNumberStrategy",
    "DocumentGenerator",
    "Accounting",
    "AllocationItem",
    "AllocationStrategy",
    "Apartment",
    "BankAccount",
    "Building",
    "Contact",
    "Invoice",
    "InvoiceCollection",
    "LaborCostItem",
    "Tenant",
    "TimePeriod",
]


from .calculations import (
    AccountingProcessor,
    ApartmentStrategy,
    DirectInvoiceStrategy,
    FloorAreaStrategy,
    InvoiceCollectionProcessor,
    ResidentNumberStrategy,
)
from .document import DocumentGenerator
from .models import (
    Accounting,
    AllocationItem,
    AllocationStrategy,
    Apartment,
    BankAccount,
    Building,
    Contact,
    Invoice,
    InvoiceCollection,
    LaborCostItem,
    Tenant,
    TimePeriod,
)



from dataclasses import dataclass

from .bank_account import BankAccount


@dataclass(frozen=True, slots=True)
class Contact:
    name: str
    street: str
    house_number: str
    zip_code: str
    city: str
    bank_account: BankAccount | None = None



from dataclasses import dataclass
from decimal import Decimal
from typing import Any

from nekore.models import Building, Tenant, TimePeriod


@dataclass(frozen=True, slots=True)
class ApartmentStrategy:
    name: str = "STK"

    def total_shares(self, _: TimePeriod, building: Building) -> Decimal:
        return Decimal(len(building.apartments))

    def tenant_shares(
        self, accounting_period: TimePeriod, _: Any, tenant: Tenant
    ) -> Decimal:
        days_of_use = Decimal(tenant.time_of_use_in(accounting_period).days)
        return days_of_use / accounting_period.duration.days



from typing import Final

__all__: Final = [
    "AccountingProcessor",
    "ApartmentStrategy",
    "DirectInvoiceStrategy",
    "FloorAreaStrategy",
    "ResidentNumberStrategy",
    "InvoiceCollectionProcessor",
]


from .accounting_processor import AccountingProcessor
from .apartment_strategy import ApartmentStrategy
from .direct_invoice_strategy import DirectInvoiceStrategy
from .floor_area_strategy import FloorAreaStrategy
from .invoice_collection_processor import InvoiceCollectionProcessor
from .resident_number_strategy import ResidentNumberStrategy



from dataclasses import dataclass, field
from decimal import Decimal
from typing import Collection

from .tenant import Tenant, overlapping_in_time


@dataclass(frozen=True, eq=True, slots=True)
class Apartment:
    floor_space: Decimal
    name: str | None = None
    tenants: Collection[Tenant] = field(default_factory=tuple, compare=False)

    def __post_init__(self) -> None:
        self._validate_tenant_periods()

    def _validate_tenant_periods(self) -> None:
        first = next(overlapping_in_time(self.tenants), None)
        if first:
            raise ValueError(f"{first[0]} and {first[1]} have overlapping periods.")
